@@ -30,31 +30,44 @@ class LayoutSchema(object):
 
     manager = v.Any('IndependentPipelineManager',
                     'DependentPipelineManager')
+
+    precedence = v.Any('normal', 'low', 'high')
+
     variable_dict = v.Schema({}, extra=True)
 
-    trigger = {v.Required('event'): toList(v.Any('patchset-created',
-                                                 'change-abandoned',
-                                                 'change-restored',
-                                                 'change-merged',
-                                                 'comment-added',
-                                                 'ref-updated')),
-               'comment_filter': toList(str),
-               'email_filter': toList(str),
-               'branch': toList(str),
-               'ref': toList(str),
-               'approval': toList(variable_dict),
-               }
+    gerrit_trigger = {v.Required('event'):
+                      toList(v.Any('patchset-created',
+                                   'change-abandoned',
+                                   'change-restored',
+                                   'change-merged',
+                                   'comment-added',
+                                   'ref-updated')),
+                      'comment_filter': toList(str),
+                      'email_filter': toList(str),
+                      'branch': toList(str),
+                      'ref': toList(str),
+                      'approval': toList(variable_dict),
+                      }
+
+    timer_trigger = {v.Required('time'): str}
+
+    trigger = v.Required(v.Any({'gerrit': toList(gerrit_trigger)},
+                               {'timer': toList(timer_trigger)}))
+
+    report_actions = {'gerrit': variable_dict,
+                      'smtp': variable_dict}
 
     pipeline = {v.Required('name'): str,
                 v.Required('manager'): manager,
+                'precedence': precedence,
                 'description': str,
                 'success-message': str,
                 'failure-message': str,
                 'dequeue-on-new-patchset': bool,
-                'trigger': toList(trigger),
-                'success': variable_dict,
-                'failure': variable_dict,
-                'start': variable_dict,
+                'trigger': trigger,
+                'success': report_actions,
+                'failure': report_actions,
+                'start': report_actions,
                 }
     pipelines = [pipeline]
 
@@ -84,7 +97,7 @@ class LayoutSchema(object):
             for k, v in value.items():
                 self.validateJob(v, path + [k])
         else:
-            self.job_name.validate(path, self.job_name.schema, value)
+            self.job_name.schema(value)
 
     def validateTemplateCalls(self, calls):
         """ Verify a project pass the parameters required
@@ -150,7 +163,8 @@ class LayoutSchema(object):
             self.templates_schemas[t_name] = v.Schema(schema)
 
         project = {'name': str,
-                   'merge-mode': v.Any('cherry-pick'),
+                   'merge-mode': v.Any('merge', 'merge-resolve,',
+                                       'cherry-pick'),
                    'template': self.validateTemplateCalls,
                    }
 
