@@ -15,12 +15,32 @@
 import re
 import time
 from uuid import uuid4
+import extras
+
+OrderedDict = extras.try_imports(['collections.OrderedDict',
+                                  'ordereddict.OrderedDict'])
 
 
-FAST_FORWARD_ONLY = 1
-MERGE_ALWAYS = 2
-MERGE_IF_NECESSARY = 3
-CHERRY_PICK = 4
+MERGER_MERGE = 1          # "git merge"
+MERGER_MERGE_RESOLVE = 2  # "git merge -s resolve"
+MERGER_CHERRY_PICK = 3    # "git cherry-pick"
+
+MERGER_MAP = {
+    'merge': MERGER_MERGE,
+    'merge-resolve': MERGER_MERGE_RESOLVE,
+    'cherry-pick': MERGER_CHERRY_PICK,
+}
+
+PRECEDENCE_NORMAL = 0
+PRECEDENCE_LOW = 1
+PRECEDENCE_HIGH = 2
+
+PRECEDENCE_MAP = {
+    None: PRECEDENCE_NORMAL,
+    'low': PRECEDENCE_LOW,
+    'normal': PRECEDENCE_NORMAL,
+    'high': PRECEDENCE_HIGH,
+}
 
 
 class Pipeline(object):
@@ -34,6 +54,11 @@ class Pipeline(object):
         self.job_trees = {}  # project -> JobTree
         self.manager = None
         self.queues = []
+        self.precedence = PRECEDENCE_NORMAL
+        self.trigger = None
+        self.start_actions = None
+        self.success_actions = None
+        self.failure_actions = None
 
     def __repr__(self):
         return '<Pipeline %s>' % self.name
@@ -47,7 +72,7 @@ class Pipeline(object):
         return job_tree
 
     def getProjects(self):
-        return self.job_trees.keys()
+        return sorted(self.job_trees.keys(), lambda a, b: cmp(a.name, b.name))
 
     def addQueue(self, queue):
         self.queues.append(queue)
@@ -68,20 +93,20 @@ class Pipeline(object):
             return []
         return changeish.filterJobs(tree.getJobs())
 
-    def _findJobsToRun(self, job_trees, changeish):
+    def _findJobsToRun(self, job_trees, item):
         torun = []
-        if changeish.change_ahead:
+        if item.item_ahead:
             # Only run jobs if any 'hold' jobs on the change ahead
             # have completed successfully.
-            if self.isHoldingFollowingChanges(changeish.change_ahead):
+            if self.isHoldingFollowingChanges(item.item_ahead):
                 return []
         for tree in job_trees:
             job = tree.job
             result = None
             if job:
-                if not job.changeMatches(changeish):
+                if not job.changeMatches(item.change):
                     continue
-                build = changeish.current_build_set.getBuild(job.name)
+                build = item.current_build_set.getBuild(job.name)
                 if build:
                     result = build.result
                 else:
@@ -91,93 +116,103 @@ class Pipeline(object):
             # If there is no job, this is a null job tree, and we should
             # run all of its jobs.
             if result == 'SUCCESS' or not job:
-                torun.extend(self._findJobsToRun(tree.job_trees, changeish))
+                torun.extend(self._findJobsToRun(tree.job_trees, item))
         return torun
 
-    def findJobsToRun(self, changeish):
-        tree = self.getJobTree(changeish.project)
+    def findJobsToRun(self, item):
+        tree = self.getJobTree(item.change.project)
         if not tree:
             return []
-        return self._findJobsToRun(tree.job_trees, changeish)
+        return self._findJobsToRun(tree.job_trees, item)
 
-    def areAllJobsComplete(self, changeish):
-        for job in self.getJobs(changeish):
-            build = changeish.current_build_set.getBuild(job.name)
+    def haveAllJobsStarted(self, item):
+        for job in self.getJobs(item.change):
+            build = item.current_build_set.getBuild(job.name)
+            if not build or not build.start_time:
+                return False
+        return True
+
+    def areAllJobsComplete(self, item):
+        for job in self.getJobs(item.change):
+            build = item.current_build_set.getBuild(job.name)
             if not build or not build.result:
                 return False
         return True
 
-    def didAllJobsSucceed(self, changeish):
-        for job in self.getJobs(changeish):
+    def didAllJobsSucceed(self, item):
+        for job in self.getJobs(item.change):
             if not job.voting:
                 continue
-            build = changeish.current_build_set.getBuild(job.name)
+            build = item.current_build_set.getBuild(job.name)
             if not build:
                 return False
             if build.result != 'SUCCESS':
                 return False
         return True
 
-    def didAnyJobFail(self, changeish):
-        for job in self.getJobs(changeish):
+    def didAnyJobFail(self, item):
+        for job in self.getJobs(item.change):
             if not job.voting:
                 continue
-            build = changeish.current_build_set.getBuild(job.name)
+            build = item.current_build_set.getBuild(job.name)
             if build and build.result and (build.result != 'SUCCESS'):
                 return True
         return False
 
-    def isHoldingFollowingChanges(self, changeish):
-        for job in self.getJobs(changeish):
+    def isHoldingFollowingChanges(self, item):
+        for job in self.getJobs(item.change):
             if not job.hold_following_changes:
                 continue
-            build = changeish.current_build_set.getBuild(job.name)
+            build = item.current_build_set.getBuild(job.name)
             if not build:
                 return True
             if build.result != 'SUCCESS':
                 return True
-        if not changeish.change_ahead:
-            return False
-        return self.isHoldingFollowingChanges(changeish.change_ahead)
 
-    def setResult(self, changeish, build):
-        if build.result != 'SUCCESS':
+        if not item.item_ahead:
+            return False
+        return self.isHoldingFollowingChanges(item.item_ahead)
+
+    def setResult(self, item, build):
+        if build.retry:
+            item.removeBuild(build)
+        elif build.result != 'SUCCESS':
             # Get a JobTree from a Job so we can find only its dependent jobs
-            root = self.getJobTree(changeish.project)
+            root = self.getJobTree(item.change.project)
             tree = root.getJobTreeForJob(build.job)
             for job in tree.getJobs():
                 fakebuild = Build(job, None)
                 fakebuild.result = 'SKIPPED'
-                changeish.addBuild(fakebuild)
+                item.addBuild(fakebuild)
 
-    def setUnableToMerge(self, changeish):
-        changeish.current_build_set.unable_to_merge = True
-        root = self.getJobTree(changeish.project)
+    def setUnableToMerge(self, item, msg):
+        item.current_build_set.unable_to_merge = True
+        item.current_build_set.unable_to_merge_message = msg
+        root = self.getJobTree(item.change.project)
         for job in root.getJobs():
             fakebuild = Build(job, None)
             fakebuild.result = 'SKIPPED'
-            changeish.addBuild(fakebuild)
+            item.addBuild(fakebuild)
 
-    def setDequeuedNeedingChange(self, changeish):
-        changeish.dequeued_needing_change = True
-        root = self.getJobTree(changeish.project)
+    def setDequeuedNeedingChange(self, item):
+        item.dequeued_needing_change = True
+        root = self.getJobTree(item.change.project)
         for job in root.getJobs():
             fakebuild = Build(job, None)
             fakebuild.result = 'SKIPPED'
-            changeish.addBuild(fakebuild)
+            item.addBuild(fakebuild)
 
     def getChangesInQueue(self):
         changes = []
         for shared_queue in self.queues:
-            changes.extend(shared_queue.queue)
+            changes.extend([x.change for x in shared_queue.queue])
         return changes
 
-    def getAllChanges(self):
-        changes = []
+    def getAllItems(self):
+        items = []
         for shared_queue in self.queues:
-            changes.extend(shared_queue.queue)
-            changes.extend(shared_queue.severed_heads)
-        return changes
+            items.extend(shared_queue.queue)
+        return items
 
     def formatStatusHTML(self):
         ret = ''
@@ -186,8 +221,8 @@ class Pipeline(object):
                 s = 'Change queue: %s' % queue.name
                 ret += s + '\n'
                 ret += '-' * len(s) + '\n'
-            for head in queue.getHeads():
-                ret += self.formatStatus(head, html=True)
+            for item in queue.queue:
+                ret += self.formatStatus(item, html=True)
         return ret
 
     def formatStatusJSON(self):
@@ -199,16 +234,26 @@ class Pipeline(object):
             j_queue = dict(name=queue.name)
             j_queues.append(j_queue)
             j_queue['heads'] = []
-            for head in queue.getHeads():
-                j_changes = []
-                c = head
-                while c:
-                    j_changes.append(self.formatChangeJSON(c))
-                    c = c.change_behind
+
+            j_changes = []
+            for e in queue.queue:
+                if not e.item_ahead:
+                    if j_changes:
+                        j_queue['heads'].append(j_changes)
+                    j_changes = []
+                j_changes.append(self.formatItemJSON(e))
+                if (len(j_changes) > 1 and
+                    (j_changes[-2]['remaining_time'] is not None) and
+                    (j_changes[-1]['remaining_time'] is not None)):
+                    j_changes[-1]['remaining_time'] = max(
+                        j_changes[-2]['remaining_time'],
+                        j_changes[-1]['remaining_time'])
+            if j_changes:
                 j_queue['heads'].append(j_changes)
         return j_pipeline
 
-    def formatStatus(self, changeish, indent=0, html=False):
+    def formatStatus(self, item, indent=0, html=False):
+        changeish = item.change
         indent_str = ' ' * indent
         ret = ''
         if html and hasattr(changeish, 'url') and changeish.url is not None:
@@ -218,11 +263,13 @@ class Pipeline(object):
                 changeish.url,
                 changeish._id())
         else:
-            ret += '%sProject %s change %s\n' % (indent_str,
-                                                 changeish.project.name,
-                                                 changeish._id())
+            ret += '%sProject %s change %s based on %s\n' % (
+                indent_str,
+                changeish.project.name,
+                changeish._id(),
+                item.item_ahead)
         for job in self.getJobs(changeish):
-            build = changeish.current_build_set.getBuild(job.name)
+            build = item.current_build_set.getBuild(job.name)
             if build:
                 result = build.result
             else:
@@ -241,38 +288,87 @@ class Pipeline(object):
                     job_name = '<a href="%s">%s</a>' % (url, job_name)
             ret += '%s  %s: %s%s' % (indent_str, job_name, result, voting)
             ret += '\n'
-        if changeish.change_behind:
-            ret += '%sFollowed by:\n' % (indent_str)
-            ret += self.formatStatus(changeish.change_behind, indent + 2, html)
         return ret
 
-    def formatChangeJSON(self, changeish):
+    def formatItemJSON(self, item):
+        changeish = item.change
         ret = {}
         if hasattr(changeish, 'url') and changeish.url is not None:
             ret['url'] = changeish.url
         else:
             ret['url'] = None
         ret['id'] = changeish._id()
+        if item.item_ahead:
+            ret['item_ahead'] = item.item_ahead.change._id()
+        else:
+            ret['item_ahead'] = None
+        ret['items_behind'] = [i.change._id() for i in item.items_behind]
+        ret['failing_reasons'] = item.current_build_set.failing_reasons
+        ret['zuul_ref'] = item.current_build_set.ref
         ret['project'] = changeish.project.name
-        ret['enqueue_time'] = int(changeish.enqueue_time * 1000)
+        ret['enqueue_time'] = int(item.enqueue_time * 1000)
         ret['jobs'] = []
+        max_remaining = 0
         for job in self.getJobs(changeish):
-            build = changeish.current_build_set.getBuild(job.name)
+            now = time.time()
+            build = item.current_build_set.getBuild(job.name)
+            elapsed = None
+            remaining = None
+            result = None
+            url = None
             if build:
                 result = build.result
                 url = build.url
-            else:
-                result = None
-                url = None
+                if build.start_time:
+                    if build.end_time:
+                        elapsed = int((build.end_time -
+                                       build.start_time) * 1000)
+                        remaining = 0
+                    else:
+                        elapsed = int((now - build.start_time) * 1000)
+                        if build.estimated_time:
+                            remaining = max(
+                                int(build.estimated_time * 1000) - elapsed,
+                                0)
+            if remaining and remaining > max_remaining:
+                max_remaining = remaining
             ret['jobs'].append(
                 dict(
                     name=job.name,
+                    elapsed_time=elapsed,
+                    remaining_time=remaining,
                     url=url,
                     result=result,
                     voting=job.voting))
+        if self.haveAllJobsStarted(item):
+            ret['remaining_time'] = max_remaining
+        else:
+            ret['remaining_time'] = None
         return ret
 
 
+class ActionReporter(object):
+    """An ActionReporter has a reporter and its configured paramaters"""
+
+    def __repr__(self):
+        return '<ActionReporter %s, %s>' % (self.reporter, self.params)
+
+    def __init__(self, reporter, params):
+        self.reporter = reporter
+        self.params = params
+
+    def report(self, change, message):
+        """Sends the built message off to the configured reporter.
+        Takes the change and message and adds the configured parameters.
+        """
+        return self.reporter.report(change, message, self.params)
+
+    def getSubmitAllowNeeds(self):
+        """Gets the submit allow needs from the reporter based off the
+        parameters."""
+        return self.reporter.getSubmitAllowNeeds(self.params)
+
+
 class ChangeQueue(object):
     """DependentPipelines have multiple parallel queues shared by
     different projects; this is one of them.  For instance, there may
@@ -285,7 +381,6 @@ class ChangeQueue(object):
         self.projects = []
         self._jobs = set()
         self.queue = []
-        self.severed_heads = []
         self.dependent = dependent
 
     def __repr__(self):
@@ -303,53 +398,58 @@ class ChangeQueue(object):
             self._jobs |= set(self.pipeline.getJobTree(project).getJobs())
 
     def enqueueChange(self, change):
+        item = QueueItem(self.pipeline, change)
+        self.enqueueItem(item)
+        item.enqueue_time = time.time()
+        return item
+
+    def enqueueItem(self, item):
         if self.dependent and self.queue:
-            change.change_ahead = self.queue[-1]
-            change.change_ahead.change_behind = change
-        self.queue.append(change)
-        change.enqueue_time = time.time()
+            item.item_ahead = self.queue[-1]
+            item.item_ahead.items_behind.append(item)
+        self.queue.append(item)
 
-    def dequeueChange(self, change):
-        if change in self.queue:
-            self.queue.remove(change)
-        if change in self.severed_heads:
-            self.severed_heads.remove(change)
-        if change.change_ahead:
-            change.change_ahead.change_behind = change.change_behind
-        if change.change_behind:
-            change.change_behind.change_ahead = change.change_ahead
-        change.change_ahead = None
-        change.change_behind = None
-        change.dequeue_time = time.time()
+    def dequeueItem(self, item):
+        if item in self.queue:
+            self.queue.remove(item)
+        if item.item_ahead:
+            item.item_ahead.items_behind.remove(item)
+        for item_behind in item.items_behind:
+            if item.item_ahead:
+                item.item_ahead.items_behind.append(item_behind)
+            item_behind.item_ahead = item.item_ahead
+        item.item_ahead = None
+        item.items_behind = []
+        item.dequeue_time = time.time()
 
-    def addSeveredHead(self, change):
-        self.severed_heads.append(change)
+    def moveItem(self, item, item_ahead):
+        if not self.dependent:
+            return False
+        if item.item_ahead == item_ahead:
+            return False
+        # Remove from current location
+        if item.item_ahead:
+            item.item_ahead.items_behind.remove(item)
+        for item_behind in item.items_behind:
+            if item.item_ahead:
+                item.item_ahead.items_behind.append(item_behind)
+            item_behind.item_ahead = item.item_ahead
+        # Add to new location
+        item.item_ahead = item_ahead
+        item.items_behind = []
+        if item.item_ahead:
+            item.item_ahead.items_behind.append(item)
+        return True
 
     def mergeChangeQueue(self, other):
         for project in other.projects:
             self.addProject(project)
 
-    def getHead(self):
-        if not self.queue:
-            return None
-        return self.queue[0]
-
-    def getHeads(self):
-        heads = []
-        if self.dependent:
-            h = self.getHead()
-            if h:
-                heads.append(h)
-        else:
-            heads.extend(self.queue)
-        heads.extend(self.severed_heads)
-        return heads
-
 
 class Project(object):
     def __init__(self, name):
         self.name = name
-        self.merge_mode = MERGE_IF_NECESSARY
+        self.merge_mode = MERGER_MERGE_RESOLVE
 
     def __str__(self):
         return self.name
@@ -381,17 +481,24 @@ class Job(object):
         return '<Job %s>' % (self.name)
 
     def copy(self, other):
-        self.failure_message = other.failure_message
-        self.success_message = other.success_message
-        self.failure_pattern = other.failure_pattern
-        self.success_pattern = other.success_pattern
-        self.parameter_function = other.parameter_function
+        if other.failure_message:
+            self.failure_message = other.failure_message
+        if other.success_message:
+            self.success_message = other.success_message
+        if other.failure_pattern:
+            self.failure_pattern = other.failure_pattern
+        if other.success_pattern:
+            self.success_pattern = other.success_pattern
+        if other.parameter_function:
+            self.parameter_function = other.parameter_function
+        if other.branches:
+            self.branches = other.branches[:]
+            self._branches = other._branches[:]
+        if other.files:
+            self.files = other.files[:]
+            self._files = other._files[:]
         self.hold_following_changes = other.hold_following_changes
         self.voting = other.voting
-        self.branches = other.branches[:]
-        self._branches = other._branches[:]
-        self.files = other.files[:]
-        self._files = other._files[:]
 
     def changeMatches(self, change):
         matches_branch = False
@@ -425,10 +532,9 @@ class JobTree(object):
         self.job_trees = []
 
     def addJob(self, job):
-        if job not in [x.job for x in self.job_trees]:
-            t = JobTree(job)
-            self.job_trees.append(t)
-            return t
+        t = JobTree(job)
+        self.job_trees.append(t)
+        return t
 
     def getJobs(self):
         jobs = []
@@ -458,14 +564,19 @@ class Build(object):
         self.launch_time = time.time()
         self.start_time = None
         self.end_time = None
+        self.estimated_time = None
+        self.pipeline = None
+        self.canceled = False
+        self.retry = False
+        self.parameters = {}
 
     def __repr__(self):
         return '<Build %s of %s>' % (self.uuid, self.job.name)
 
 
 class BuildSet(object):
-    def __init__(self, change):
-        self.change = change
+    def __init__(self, item):
+        self.item = item
         self.other_changes = []
         self.builds = {}
         self.result = None
@@ -474,16 +585,18 @@ class BuildSet(object):
         self.ref = None
         self.commit = None
         self.unable_to_merge = False
+        self.unable_to_merge_message = None
+        self.failing_reasons = []
 
     def setConfiguration(self):
         # The change isn't enqueued until after it's created
         # so we don't know what the other changes ahead will be
         # until jobs start.
         if not self.other_changes:
-            next_change = self.change.change_ahead
-            while next_change:
-                self.other_changes.append(next_change)
-                next_change = next_change.change_ahead
+            next_item = self.item.item_ahead
+            while next_item:
+                self.other_changes.append(next_item.change)
+                next_item = next_item.item_ahead
         if not self.ref:
             self.ref = 'Z' + uuid4().hex
 
@@ -491,6 +604,9 @@ class BuildSet(object):
         self.builds[build.job.name] = build
         build.build_set = self
 
+    def removeBuild(self, build):
+        del self.builds[build.job.name]
+
     def getBuild(self, job_name):
         return self.builds.get(job_name)
 
@@ -500,29 +616,29 @@ class BuildSet(object):
         return [self.builds.get(x) for x in keys]
 
 
-class Changeish(object):
-    """Something like a change; either a change or a ref"""
-    is_reportable = False
+class QueueItem(object):
+    """A changish inside of a Pipeline queue"""
 
-    def __init__(self, project):
-        self.project = project
+    def __init__(self, pipeline, change):
+        self.pipeline = pipeline
+        self.change = change  # a changeish
         self.build_sets = []
         self.dequeued_needing_change = False
         self.current_build_set = BuildSet(self)
         self.build_sets.append(self.current_build_set)
-        self.change_ahead = None
-        self.change_behind = None
+        self.item_ahead = None
+        self.items_behind = []
         self.enqueue_time = None
         self.dequeue_time = None
+        self.reported = False
 
-    def equals(self, other):
-        raise NotImplementedError()
-
-    def isUpdateOf(self, other):
-        raise NotImplementedError()
-
-    def filterJobs(self, jobs):
-        return filter(lambda job: job.changeMatches(self), jobs)
+    def __repr__(self):
+        if self.pipeline:
+            pipeline = self.pipeline.name
+        else:
+            pipeline = None
+        return '<QueueItem 0x%x for %s in %s>' % (
+            id(self), self.change, pipeline)
 
     def resetAllBuilds(self):
         old = self.current_build_set
@@ -534,6 +650,33 @@ class Changeish(object):
 
     def addBuild(self, build):
         self.current_build_set.addBuild(build)
+        build.pipeline = self.pipeline
+
+    def removeBuild(self, build):
+        self.current_build_set.removeBuild(build)
+
+    def setReportedResult(self, result):
+        self.current_build_set.result = result
+
+
+class Changeish(object):
+    """Something like a change; either a change or a ref"""
+    is_reportable = False
+
+    def __init__(self, project):
+        self.project = project
+
+    def equals(self, other):
+        raise NotImplementedError()
+
+    def isUpdateOf(self, other):
+        raise NotImplementedError()
+
+    def filterJobs(self, jobs):
+        return filter(lambda job: job.changeMatches(self), jobs)
+
+    def getRelatedChanges(self):
+        return set()
 
 
 class Change(Changeish):
@@ -548,12 +691,12 @@ class Change(Changeish):
         self.refspec = None
 
         self.files = []
-        self.reported = False
         self.needs_change = None
         self.needed_by_changes = []
         self.is_current_patchset = True
         self.can_merge = False
         self.is_merged = False
+        self.failed_to_merge = False
 
     def _id(self):
         return '%s,%s' % (self.number, self.patchset)
@@ -568,12 +711,21 @@ class Change(Changeish):
 
     def isUpdateOf(self, other):
         if ((hasattr(other, 'number') and self.number == other.number) and
-            (hasattr(other, 'patchset') and self.patchset > other.patchset)):
+            (hasattr(other, 'patchset') and
+             self.patchset is not None and
+             other.patchset is not None and
+             int(self.patchset) > int(other.patchset))):
             return True
         return False
 
-    def setReportedResult(self, result):
-        self.current_build_set.result = result
+    def getRelatedChanges(self):
+        related = set()
+        if self.needs_change:
+            related.add(self.needs_change)
+        for c in self.needed_by_changes:
+            related.add(c)
+            related.update(c.getRelatedChanges())
+        return related
 
 
 class Ref(Changeish):
@@ -614,12 +766,26 @@ class Ref(Changeish):
         return False
 
 
+class NullChange(Changeish):
+    is_reportable = False
+
+    def _id(self):
+        return None
+
+    def equals(self, other):
+        return False
+
+    def isUpdateOf(self, other):
+        return False
+
+
 class TriggerEvent(object):
     def __init__(self):
         self.data = None
         # common
         self.type = None
         self.project_name = None
+        self.trigger_name = None
         # Representation of the user account that performed the event.
         self.account = None
         # patchset-created, comment-added, etc.
@@ -634,6 +800,11 @@ class TriggerEvent(object):
         self.ref = None
         self.oldrev = None
         self.newrev = None
+        # timer
+        self.timespec = None
+        # For events that arrive with a destination pipeline (eg, from
+        # an admin command, etc):
+        self.forced_pipeline = None
 
     def __repr__(self):
         ret = '<TriggerEvent %s %s' % (self.type, self.project_name)
@@ -650,25 +821,23 @@ class TriggerEvent(object):
         return ret
 
     def getChange(self, project, trigger):
-        # TODO: make the scheduler deal with events (which may have
-        # changes) rather than changes so that we don't have to create
-        # "fake" changes for events that aren't associated with changes.
-
         if self.change_number:
             change = trigger.getChange(self.change_number, self.patch_number)
-        if self.ref:
+        elif self.ref:
             change = Ref(project)
             change.ref = self.ref
             change.oldrev = self.oldrev
             change.newrev = self.newrev
             change.url = trigger.getGitwebUrl(project, sha=self.newrev)
+        else:
+            change = NullChange(project)
 
         return change
 
 
 class EventFilter(object):
     def __init__(self, types=[], branches=[], refs=[], approvals={},
-                 comment_filters=[], email_filters=[]):
+                 comment_filters=[], email_filters=[], timespecs=[]):
         self._types = types
         self._branches = branches
         self._refs = refs
@@ -680,6 +849,7 @@ class EventFilter(object):
         self.comment_filters = [re.compile(x) for x in comment_filters]
         self.email_filters = [re.compile(x) for x in email_filters]
         self.approvals = approvals
+        self.timespecs = timespecs
 
     def __repr__(self):
         ret = '<EventFilter'
@@ -697,6 +867,8 @@ class EventFilter(object):
             ret += ' comment_filters: %s' % ', '.join(self._comment_filters)
         if self._email_filters:
             ret += ' email_filters: %s' % ', '.join(self._email_filters)
+        if self.timespecs:
+            ret += ' timespecs: %s' % ', '.join(self.timespecs)
         ret += '>'
 
         return ret
@@ -761,4 +933,37 @@ class EventFilter(object):
                     matches_approval = True
             if not matches_approval:
                 return False
+
+        # timespecs are ORed
+        matches_timespec = False
+        for timespec in self.timespecs:
+            if (event.timespec == timespec):
+                matches_timespec = True
+        if self.timespecs and not matches_timespec:
+            return False
+
         return True
+
+
+class Layout(object):
+    def __init__(self):
+        self.projects = {}
+        self.pipelines = OrderedDict()
+        self.jobs = {}
+        self.metajobs = []
+
+    def getJob(self, name):
+        if name in self.jobs:
+            return self.jobs[name]
+        job = Job(name)
+        if name.startswith('^'):
+            # This is a meta-job
+            regex = re.compile(name)
+            self.metajobs.append((regex, job))
+        else:
+            # Apply attributes from matching meta-jobs
+            for regex, metajob in self.metajobs:
+                if regex.match(name):
+                    job.copy(metajob)
+            self.jobs[name] = job
+        return job
