@@ -25,10 +25,12 @@ class GerritEventConnector(threading.Thread):
 
     log = logging.getLogger("zuul.GerritEventConnector")
 
-    def __init__(self, gerrit, sched):
+    def __init__(self, gerrit, sched, trigger):
         super(GerritEventConnector, self).__init__()
+        self.daemon = True
         self.gerrit = gerrit
         self.sched = sched
+        self.trigger = trigger
         self._stopped = False
 
     def stop(self):
@@ -41,6 +43,7 @@ class GerritEventConnector(threading.Thread):
             return
         event = TriggerEvent()
         event.type = data.get('type')
+        event.trigger_name = self.trigger.name
         change = data.get('change')
         if change:
             event.project_name = change.get('project')
@@ -80,6 +83,14 @@ class GerritEventConnector(threading.Thread):
                     Can not get account information." % event.type)
             event.account = None
 
+        if event.change_number:
+            # Call getChange for the side effect of updating the
+            # cache.  Note that this modifies Change objects outside
+            # the main thread.
+            self.trigger.getChange(event.change_number,
+                                   event.patch_number,
+                                   refresh=True)
+
         self.sched.addEvent(event)
         self.gerrit.eventDone()
 
@@ -94,11 +105,13 @@ class GerritEventConnector(threading.Thread):
 
 
 class Gerrit(object):
+    name = 'gerrit'
     log = logging.getLogger("zuul.Gerrit")
     replication_timeout = 60
     replication_retry_interval = 5
 
     def __init__(self, config, sched):
+        self._change_cache = {}
         self.sched = sched
         self.config = config
         self.server = config.get('gerrit', 'server')
@@ -118,29 +131,13 @@ class Gerrit(object):
         self.gerrit = gerrit.Gerrit(self.server, user, port, sshkey)
         self.gerrit.startWatching()
         self.gerrit_connector = GerritEventConnector(
-            self.gerrit, sched)
+            self.gerrit, sched, self)
         self.gerrit_connector.start()
 
     def stop(self):
         self.gerrit_connector.stop()
         self.gerrit_connector.join()
 
-    def report(self, change, message, action):
-        self.log.debug("Report change %s, action %s, message: %s" %
-                       (change, action, message))
-        if not change.number:
-            self.log.debug("Change has no number; not reporting")
-            return
-        if not action:
-            self.log.debug("No action specified; not reporting")
-            return
-        changeid = '%s,%s' % (change.number, change.patchset)
-        ref = 'refs/heads/' + change.branch
-        change._ref_sha = self.getRefSha(change.project.name,
-                                         ref)
-        return self.gerrit.review(change.project.name, changeid,
-                                  message, action)
-
     def _getInfoRefs(self, project):
         url = "%s/p/%s/info/refs?service=git-upload-pack" % (
             self.baseurl, project)
@@ -278,29 +275,65 @@ class Gerrit(object):
             return False
         return True
 
-    def getChange(self, number, patchset, changes=None):
-        self.log.info("Getting information for %s,%s" % (number, patchset))
-        if changes is None:
-            changes = {}
-        data = self.gerrit.query(number)
-        project = self.sched.projects[data['project']]
-        change = Change(project)
+    def maintainCache(self, relevant):
+        # This lets the user supply a list of change objects that are
+        # still in use.  Anything in our cache that isn't in the supplied
+        # list should be same to remove from the cache.
+        remove = []
+        for key, change in self._change_cache.items():
+            if change not in relevant:
+                remove.append(key)
+        for key in remove:
+            del self._change_cache[key]
+
+    def postConfig(self):
+        pass
+
+    def getChange(self, number, patchset, refresh=False):
+        key = '%s,%s' % (number, patchset)
+        change = None
+        if key in self._change_cache:
+            change = self._change_cache.get(key)
+            if not refresh:
+                return change
+        if not change:
+            change = Change(None)
+            change.number = number
+            change.patchset = patchset
+        key = '%s,%s' % (change.number, change.patchset)
+        self._change_cache[key] = change
+        try:
+            self.updateChange(change)
+        except Exception:
+            del self._change_cache[key]
+            raise
+        return change
+
+    def updateChange(self, change):
+        self.log.info("Updating information for %s,%s" %
+                      (change.number, change.patchset))
+        data = self.gerrit.query(change.number)
         change._data = data
 
-        change.number = number
-        change.patchset = patchset
-        change.project = project
+        if change.patchset is None:
+            change.patchset = data['currentPatchSet']['number']
+
+        if 'project' not in data:
+            raise Exception("Change %s,%s not found" % (change.number,
+                                                        change.patchset))
+        change.project = self.sched.getProject(data['project'])
         change.branch = data['branch']
         change.url = data['url']
         max_ps = 0
+        change.files = []
         for ps in data['patchSets']:
-            if ps['number'] == patchset:
+            if ps['number'] == change.patchset:
                 change.refspec = ps['ref']
                 for f in ps.get('files', []):
                     change.files.append(f['file'])
             if int(ps['number']) > int(max_ps):
                 max_ps = ps['number']
-        if max_ps == patchset:
+        if max_ps == change.patchset:
             change.is_current_patchset = True
         else:
             change.is_current_patchset = False
@@ -311,28 +344,20 @@ class Gerrit(object):
             # for dependencies.
             return change
 
-        key = '%s,%s' % (number, patchset)
-        changes[key] = change
-
-        def cachedGetChange(num, ps):
-            key = '%s,%s' % (num, ps)
-            if key in changes:
-                return changes.get(key)
-            c = self.getChange(num, ps, changes)
-            return c
-
+        change.needs_change = None
         if 'dependsOn' in data:
             parts = data['dependsOn'][0]['ref'].split('/')
             dep_num, dep_ps = parts[3], parts[4]
-            dep = cachedGetChange(dep_num, dep_ps)
+            dep = self.getChange(dep_num, dep_ps)
             if not dep.is_merged:
                 change.needs_change = dep
 
+        change.needed_by_changes = []
         if 'neededBy' in data:
             for needed in data['neededBy']:
                 parts = needed['ref'].split('/')
                 dep_num, dep_ps = parts[3], parts[4]
-                dep = cachedGetChange(dep_num, dep_ps)
+                dep = self.getChange(dep_num, dep_ps)
                 if not dep.is_merged and dep.is_current_patchset:
                     change.needed_by_changes.append(dep)
 
